/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.common.artifact;

import java.io.IOException;

import org.n52.security.common.xml.DOMParser;
import org.n52.security.common.xml.XMLPathCtx;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

/**
 * This exception is thrown when an error during the was process occurred. This implementation
 * supports just service exceptions, which contain exactly one <code>&lt ServiceException&gt</code> element.
 */
public class ServiceException extends Exception implements TransferableCreator {

    private static final long serialVersionUID = 5528834878852780744L;

    public static final String AUTHENTICATION_FAILED = "AuthenticationFailed";

    public static final String AUTHORIZATION_FAILED = "AuthorizationFailed";

    public static final String PASS_FORMATION_ERROR = "PasswordFormationError";

    public static final String SESSION_EXPIRED = "SessionExpired";

    public static final String INVALID_FORMAT = "InvalidFormat";

    public static final String INVALID_SESSION = "InvalidSessionID";

    public static final String SERVICE_ERROR = "ServiceError";

    public static final String INVALID_SAMLRESPONSE = "InvalidSAMLResponse";

    private static final String MIME_TYPE = "application/vnd.ogc.se_xml";

    protected String m_code;

    /**
     * Constructs an Exception with a detailed message.
     *
     * @param message
     *        The message associated with the exception.
     * @param code
     *        Error code describing the exception. Code values are defines in the WAS and WSS specification.
     *        However, it is recommended to use the constants of this class, which completely reproduce the
     *        specified error codes.
     */
    public ServiceException(String message, String code) {
        super(message);
        this.m_code = code;
    }

    public ServiceException(String message, String code, Throwable cause){
        super(message, cause);
        this.m_code = code;
    }

    public String getErrorCode() {
        return m_code;
    }

    /**
     * Creates an instance of this class from an XML file.
     *
     * @param in
     *        a service excpetion XML document
     *
     * @return
     * @throws org.xml.sax.SAXException
     * @throws IOException
     */
    public static ServiceException createFromXML(InputSource in) {

        Document excDoc = DOMParser.createNew().parse(in);

        XMLPathCtx pathCtx = XMLPathCtx.createNew();
        String cd = pathCtx.findIn(excDoc).text("//ServiceException/@code").get().trim();

        if (cd.equals(""))
        	cd = "No code available";

        String msg = pathCtx.findIn(excDoc).text("//ServiceException/text()").get().trim();
        if (msg.equals(""))
            msg = "No exception information available";

        return new ServiceException(msg, cd);
    }

    public Transferable getAsTransferable() {

        StringBuffer buf = new StringBuffer();
        buf.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        buf.append("<ServiceExceptionReport version=\"1.1.0\">\n");
        buf.append("<ServiceException code=\"" + m_code + "\">\n");
        buf.append("<![CDATA[");
        buf.append(getMessage());
        buf.append("]]>");
        buf.append("</ServiceException>\n");
        buf.append("</ServiceExceptionReport>\n");

        return TransferableFactory.getInstance().createTextualTransferable(MIME_TYPE,buf.toString(),"UTF-8");
    }
}