/**
 * *****************************************************************************
 * Copyright (C) 2008 by 52 North Initiative for Geospatial Open Source Software
 * GmbH
 *
 * Contact: Andreas Wytzisk 52 North Initiative for Geospatial Open Source
 * Software GmbH Martin-Luther-King-Weg 24 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 * ****************************************************************************
 */
package org.n52.security.common.util;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * A collection of useful tools for servlet handling.
 */
public class ServletUtils {

    private static boolean mayContainRequestParameter(HttpServletRequest httpRequest) {
        String contentType = emptyIfNull(httpRequest.getContentType());
        return "GET".equals(httpRequest.getMethod())
                || isWWW_FORM_URLEncoded(contentType);
    }

    private static boolean isWWW_FORM_URLEncoded(String contentType) {
        // correct is "application/x-www-form-urlencoded"
        // but sometimes wrong types are given
        return contentType == null || contentType.isEmpty()
                || contentType.contains("x-www-form-urlencoded")
                || contentType.contains("x-www-form-encoded");
    }

    /**
     * Returns a parameter's value by its name ignoring the case.
     *
     * @param request the request.
     * @param paramname the paramname.
     * @return String.
     */
    public static String getParameterValueEqualIgnoreCase(final HttpServletRequest request, final String paramname) {
        if (mayContainRequestParameter(request)) {
            Enumeration names = request.getParameterNames();
            while (names.hasMoreElements()) {
                String name = (String) names.nextElement();
                if (name.equalsIgnoreCase(paramname)) {
                    return request.getParameter(name);
                }
            }
        }
        return null;
    }

    public static String buildContextLocation(HttpServletRequest request, String... contextRelativePath) {
        if (contextRelativePath == null || contextRelativePath.length == 0) {
            return getContextLocation(request);
        }
        StringBuilder path = new StringBuilder(getContextLocation(request));
        boolean first = true;
        for (String subpath : contextRelativePath) {
            if (subpath != null) {
                if (first && subpath.matches("^(?i:https?)://.*")) {
                    path.setLength(0);
                    path.append(subpath);
                } else {
                    path.append(subpath);
                }
            }
            first = false;
        }
        return path.toString();
    }

    public static String getContextLocation(HttpServletRequest request) {
        try {
            HttpServletRequest httpRequest = request;
            return URLUtils.removeDefaultPort(new URL(httpRequest.getScheme(),
                    httpRequest.getServerName(),
                    httpRequest.getServerPort(),
                    httpRequest.getContextPath()).toExternalForm());
        } catch (MalformedURLException e) {
            throw new IllegalStateException("can't create valid context url:" + e, e);
        }
    }

    /**
     * @return a servlet's URL from the request.
     *
     */
    public static String getServletLocation(final HttpServletRequest request) {
        return buildContextLocation(request, request.getServletPath());
    }

    public static String getRequestLocation(final HttpServletRequest request) {
        return buildContextLocation(request, request.getServletPath(), request.getPathInfo());
        // return URLUtils.removeDefaultPort(httpRequest.getRequestURL().toString());
    }

    /**
     * Appends all parameters from the request (also HTTP POST
     * parameters) to the request location.
     */
    public static String getRequestLocationIncludingParameter(final HttpServletRequest request) {
        return getRequestLocationIncludingParameter(request, true);
    }

    /**
     * Switches between getRequestLocationIncludingQueryParameter and
     * getRequestLocationIncludingParameter based on the
     * consumeFORMURLEncodedBody flag.
     */
    public static String getRequestLocationIncludingParameter(final HttpServletRequest request, boolean consumeFORMURLEncodedBody) {
        if (!consumeFORMURLEncodedBody || "GET".equals(request.getMethod())) {
            return getRequestLocationIncludingQueryParameter(request);
        }
        if (mayContainRequestParameter(request)) {
            StringBuilder currentRequestLocation = new StringBuilder(getRequestLocation(request));
            appendRequestParameter(currentRequestLocation, request);
            return currentRequestLocation.toString();
        }
        return getRequestLocationIncludingQueryParameter(request);
    }

    /**
     * Appends the query parameters to the request location.
     */
    public static String getRequestLocationIncludingQueryParameter(HttpServletRequest request) {
        String queryPart = request.getQueryString();
        StringBuffer requestUrl = request.getRequestURL();
        if (queryPart != null && queryPart.length() > 0) {
            requestUrl.append("?").append(queryPart);
        }
        return requestUrl.toString();
    }

    private static void appendRequestParameter(StringBuilder currentRequestLocation, HttpServletRequest httpRequest) {
        Enumeration params = httpRequest.getParameterNames();
        if (!params.hasMoreElements()) {
            return;
        }
        currentRequestLocation.append("?");
        while (params.hasMoreElements()) {
            String name = (String) params.nextElement();
            String val = httpRequest.getParameter(name);
            currentRequestLocation.append(URLUtils.encodeForURL(name));
            if (val != null) {
                currentRequestLocation.append("=").append(URLUtils.encodeForURL(val));
            }
            currentRequestLocation.append("&");
        }
        int lastIndex = currentRequestLocation.length() - 1;
        if (currentRequestLocation.charAt(lastIndex) == '&') {
            currentRequestLocation.deleteCharAt(lastIndex);
        }
    }

    /**
     * Sends a redirect not using the response.sendRedirect method to allow send
     * Set-Cookie headers.
     */
    public static void sendLocationHeaderRedirect(String redirectUrl, HttpServletResponse response) {

        DateFormat dateFormat = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss 'GMT'", Locale.ENGLISH);
        dateFormat.setTimeZone(TimeZone.getTimeZone("GMT"));


        String encodedRedirectUrl = response.encodeRedirectURL(redirectUrl);
        response.setStatus(HttpServletResponse.SC_SEE_OTHER);
        response.setHeader("Location", encodedRedirectUrl);

        // Add cache controlling headers, otherwise Safari does not follow the redirect but uses a cached response
        response.setHeader("Cache-Control", "no-cache, no-store");
        response.setHeader("Expires", dateFormat.format(new Date(0)));

    }

    /**
     * Initiates redirect based on the length of the redirection URL.
     * Chooses between location header redirect or form POST redirect.
     */
    public static void sendRedirect(String redirectUrl, HttpServletResponse response) {
        // max length of url in ie is 2.083
        if (redirectUrl.length() > 2083) {
            sendFormPostRedirect(redirectUrl, response);
            return;
        }
        sendLocationHeaderRedirect(redirectUrl, response);
    }

    /**
     * Sends a HTML form redirect. Builds an HTML fragment as content containing a form
     * which sends itself after loading. Usefull for URLs
     * with a long parameter list. All URL parameters are encoded as POST body
     * parameters.
     */
    public static void sendFormPostRedirect(String redirectUrl, HttpServletResponse response) {
        sendFormPostRedirect(redirectUrl, response, URLUtils.parseQueryStringFromUrl(redirectUrl));
    }

    public static void sendFormPostRedirect(String redirectUrl, HttpServletResponse response, Map<String, Iterable<String>> params) {
        String htmlRedirect = buildHTMLFormRedirect(redirectUrl, params);
        response.setContentType("text/html;charset=UTF-8");
        try {
            PrintWriter writer = response.getWriter();
            writer.write(htmlRedirect);
        } catch (IOException e) {
            throw new IllegalStateException("can't resolve writer from http servlet response", e);
        }
    }

    private static String buildHTMLFormRedirect(String redirectUrl, Map<String, Iterable<String>> params) {
        redirectUrl = URLUtils.removeQueryStringFromUrl(redirectUrl);
        StringBuilder formPostContent = new StringBuilder();
        formPostContent
                .append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
                        + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">"
                        + "<html><head><title>sso redirect</title>"
                        + "<meta http-equiv=\"Expires\" content=\"Mon, 01 Jan 1990 00:00:01 GMT\" />"
                        + "<meta http-equiv=\"pragma\" content=\"no-cache\" />"
                        + "<meta http-equiv=\"cache-control\" content=\"no-cache\" />"
                        + "</head>"
                        + "<body onload=\"document.forms[0].submit();\">");
        formPostContent.append("<form action=\"").append(redirectUrl).append("\" method=\"POST\">");
        for (String key : params.keySet()) {
            for (String val : params.get(key)) {
                formPostContent.append("<input type=\"hidden\" name=\"")
                        .append(key)
                        .append("\" value=\"")
                        .append(val)
                        .append("\"/>");
            }
        }
        formPostContent.append("</form></body></html>");
        return formPostContent.toString();
    }

    private static String emptyIfNull(String aString) {
        return aString == null ? "" : aString;
    }

}
