package org.n52.security.common.util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Allows a single Spring-configured Timer in an application, at which some timer definitions can be
 * registered.
 */
public class TimerTaskScheduler {

    private static final String DEFAULT_TIMER_NAME = "TimerTaskScheduler-Timer";

    private static final Logger LOG = LoggerFactory.getLogger(TimerTaskScheduler.class);

    /**
     * The single timer to use for different tasks.
     */
    private Timer m_timer;
    /**
     * flag to indicate that the timer was created by this scheduler.
     */
    private boolean m_cancelTimer = false;
    /**
     * The name of the timer, only valid if self created.
     */
    private String m_timerName = DEFAULT_TIMER_NAME;
    /**
     * The task definitions, which will be scheduled during start.
     */
    private List<TimerTaskDefinition> m_taskDefinitions = new ArrayList<TimerTaskDefinition>();
    /**
     * Flag if scheduler is started.
     */
    private boolean m_started = false;

    /**
     * Life Time Method, normally called within an Dependency Injection Container.
     * It goes through all tasks definitions and add them to the timer.
     */
    public void start() {
        initTimer();
        m_started = true;
        List<TimerTaskDefinition> defs = m_taskDefinitions;
        for (TimerTaskDefinition def : defs) {
            scheduleTaskDefinition(def);
        }
        // clear task list
        m_taskDefinitions = new ArrayList<TimerTaskDefinition>();
    }

    /**
     * Life-cycle method, normally called within an Dependency Injection Container.
     * It cancels the timer.
     */
    public void shutdown() {
        m_started = false;
        if (m_cancelTimer && m_timer != null) {
            m_timer.cancel();
        }
        // clear reference
        m_timer = null;
        m_taskDefinitions = Collections.emptyList();
    }

    /**
     * Registers a new task definition with this scheduler.
     */
    public void scheduleTaskDefinition(TimerTaskDefinition def) {
        if (def == null) {
            throw new IllegalArgumentException("<timerTaskDefinition> must not null!");
        }
        if (!m_started) {
            // queue for start
            m_taskDefinitions.add(def);
            return;
        }

        long delay = Math.max(def.getDelay(), 0);
        long period = Math.max(def.getPeriod(), -1);
        Runnable task = def.getTask();
        if (task == null) {
            throw new IllegalArgumentException("<timerTaskDefinition> has null task!");
        }

        TimerTask timerTask = task instanceof TimerTask ? (TimerTask) task : new TimerTaskRunnableAdapter(task);
        Timer timer = getTimer();
        if (LOG.isDebugEnabled()) {
            LOG.debug("register task <{}>, delay:<{}>, period:<{}>", new Object[] { def, delay, period });
        }
        if (period == -1) {
            timer.schedule(timerTask, delay);
        } else {
            timer.schedule(timerTask, delay, period);
        }
    }

    private Timer initTimer() {
        if (m_timer == null) {
            m_timer = new Timer(getTimerName(), true);
            m_cancelTimer = true;
        }
        return m_timer;
    }

    public Timer getTimer() {
        return m_timer;
    }

    public void setTimer(Timer timer) {
        m_timer = timer;
    }

    public String getTimerName() {
        String name = m_timerName;
        return name == null || name.isEmpty() ? DEFAULT_TIMER_NAME : name;
    }

    public void setTimerName(String timerName) {
        m_timerName = timerName;
    }

    public List<TimerTaskDefinition> getTaskDefinitions() {
        return Collections.unmodifiableList(m_taskDefinitions);
    }

    public void setTaskDefinitions(List<TimerTaskDefinition> taskDefinitions) {
        m_taskDefinitions = new ArrayList<TimerTaskDefinition>(taskDefinitions);
    }

}
