/*******************************************************************************
 * Copyright (C) 2008 by 52 North Initiative for Geospatial Open Source Software
 * GmbH
 *
 * Contact: Andreas Wytzisk 52 North Initiative for Geospatial Open Source
 * Software GmbH Martin-Luther-King-Weg 24 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/
package org.n52.security.common.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.zip.GZIPOutputStream;

import javax.servlet.http.HttpServletResponse;

import org.n52.security.common.artifact.HttpHeaderAttribute;
import org.n52.security.common.artifact.TransferAttribute;
import org.n52.security.common.artifact.Transferable;

/**
 * Provides methods to write {@link Transferable} instances to a Servlet's output stream.
 */
public class TransferableServletWriter {

    private static final org.slf4j.Logger LOG = org.slf4j.LoggerFactory.getLogger(TransferableServletWriter.class);
    private final Transferable m_transferable;


    public TransferableServletWriter(Transferable transferable) {
        m_transferable = transferable;
    }

    /**
     * Writes the content of the transferable to the servlet response's output stream.
     * The output HTTP header are set regarding
     *
     * <ul>
     *    <li>HttpHeaderAttribute.HEADER_NAME_CONTENTTYPE</li>
     *    <li>HttpHeaderAttribute.HEADER_NAME_COOKIE</li>
     * </ul>
     *
     * If the transferable contains an HttpHeaderAttribute.HEADER_NAME_CONTENT_ENCODING attribute and its value
     * contains 'gzip', the stream is written using gzip compression.
     *
     * @param transferable
     * @param servletResponse
     */
    public void write(HttpServletResponse servletResponse) {
        setStatusCodeHeader(servletResponse);
        setContentTypeHeader(servletResponse);
        TransferAttribute contentEncoding =
            getTransferable().getAttribute(HttpHeaderAttribute.HEADER_NAME_CONTENT_ENCODING);
        boolean gzip = contentEncoding != null && "gzip".equalsIgnoreCase((String) contentEncoding.getValue());
        if (gzip) {
            // ok no charset (maybe gzip or such a thing!)
            addHeader(servletResponse, contentEncoding);
            if (LOG.isDebugEnabled()) {
                LOG.debug("enable <gzip> compression!");
            }
        }
        // cookie support has to be revalidated!
        addHeader(servletResponse, getTransferable().getAttribute(HttpHeaderAttribute.HEADER_NAME_COOKIE));
        OutputStream destinationStream;
        try {
            destinationStream = servletResponse.getOutputStream();
        } catch (IOException e) {
            LOG.warn("Error writing stream");
            return;
        }
        InputStream sourceStream  = getTransferable().getPayload().getAsStream();
        writeStream(sourceStream, destinationStream, gzip);
    }

    public void setContentTypeHeader(HttpServletResponse servletResponse) {
        addHeader(servletResponse, getTransferable().getAttribute(HttpHeaderAttribute.HEADER_NAME_CONTENTTYPE));
    }

    public void setStatusCodeHeader(HttpServletResponse servletResponse) {
        String httpCode = (String) getTransferable().getAttributeValue(HttpHeaderAttribute.RESPONSE_STATUS_CODE);
        if (httpCode != null && httpCode.length() > 0) {
            // write the correct status code
            servletResponse.setStatus(Integer.parseInt(httpCode));
        }        // Write only known HttpHeaders!
    }

    private void addHeader(HttpServletResponse pResp, TransferAttribute pAttribute) {
        if (pAttribute == null) {
            return;
        }
        pResp.addHeader(pAttribute.getName(), (String) pAttribute.getValue());
    }

    private void writeStream(InputStream sourceStream, OutputStream destinationStream, boolean gzip) {
        OutputStream finalDestinationStream = null;
        try {
            finalDestinationStream =
                gzip ? (OutputStream) new GZIPOutputStream(destinationStream, 1024) : destinationStream;
            byte[] lBuffer = new byte[1024];
            int lLen = -1;
            while ((lLen = sourceStream.read(lBuffer)) != -1) {
                finalDestinationStream.write(lBuffer, 0, lLen);
            }
        } catch (IOException e) {
            LOG.error("Error while writing service response to stream", e);
        } finally {
            try {
                if (finalDestinationStream != null) {
                    finalDestinationStream.flush();
                    finalDestinationStream.close();
                }
            } catch (IOException e) {
            }
            try {
                if (sourceStream != null) {
                    sourceStream.close();
                }
            } catch (IOException e) {
            }
        }
    }


    public Transferable getTransferable() {
        return m_transferable;
    }


}
