/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/
package org.n52.security.common.util;

import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Utilities for URL analysis.
 */
public final class URLUtils {
    private URLUtils() {}

    public static String encodeForURL(String value) {
        return encodeForURL(value, null);
    }

    public static String encodeForURL(String value, String charset) {
        charset = charset == null ? "UTF-8" : charset;
        try {
            return URLEncoder.encode(value, charset);
        } catch (UnsupportedEncodingException e) {
            throw new IllegalStateException("Unknown encoding <" + charset + ">", e);
        }
    }
    public static String decodeFromURL(String value) {
        return decodeFromURL(value, null);
    }
    public static String decodeFromURL(String value, String charset) {
        charset = charset == null ? "UTF-8" : charset;
        try {
            return URLDecoder.decode(value, charset);
        } catch (UnsupportedEncodingException e) {
            throw new IllegalStateException("Unknown encoding <" + charset + ">", e);
        }
    }

    public static String replaceQueryInUrl(String url, Map<String, Iterable<String>> params) {
        String query = buildQueryStringForUrl(params);
        url = removeQueryStringFromUrl(url);
        return url + "?" + query;
    }

    public static String appendQueryToUrl(String url, Map<String, Iterable<String>> params) {
        return appendQueryToUrl(url, buildQueryStringForUrl(params));
    }

    public static String appendQueryToUrl(String url, String query) {
        if (query == null || query.isEmpty()) {
            return url;
        }
        if (url.contains("?")) {
            if (url.charAt(url.length() - 1) == '?') {
                return url + query;
            }
            return url + "&" + query;
        } else {
            return url + "?" + query;
        }
    }

    public static String buildQueryStringForUrl(Map<String, Iterable<String>> params) {
        return buildQueryStringForUrl(params, null);
    }

    public static String buildQueryStringForUrl(Map<String, Iterable<String>> params, String charset) {
        StringBuilder queryString = new StringBuilder();
        for (String key : params.keySet()) {
            Iterable<String> values = params.get(key);
            Iterator<String> valueIterator = values.iterator();
            if (valueIterator.hasNext()) {
                while (valueIterator.hasNext()) {
                    String value = valueIterator.next();
                    queryString.append(encodeForURL(key, charset));
                    queryString.append("=");
                    queryString.append(encodeForURL(value, charset));
                    queryString.append("&");
                }
            } else {
                // only append key
                queryString.append(encodeForURL(key, charset));
                queryString.append("&");
            }
        }
        if (queryString.length() > 0) {
            // delete ending '&'
            queryString.deleteCharAt(queryString.length() - 1);
        }
        return queryString.toString();
    }

    public static Map<String, Iterable<String>> parseQueryStringFromUrl(String url) {
        return parseQueryStringFromUrl(url, null);
    }

    public static Map<String, Iterable<String>> parseQueryStringFromUrl(String url, String charset) {
        Map<String, Iterable<String>> params = new HashMap<String, Iterable<String>>();
        String query = getQueryStringFromUrl(url);
        if (query.isEmpty()) {
            return params;
        }
        String[] kvps = query.split("&+");
        for (String kvp : kvps) {
            String key;
            String val;
            int indexOfEqual = kvp.indexOf('=');
            if (indexOfEqual < 0) {
                key = kvp;
                val = "";
            } else {
                key = kvp.substring(0, indexOfEqual);
                val = kvp.substring(indexOfEqual + 1);
            }
            key = decodeFromURL(key, charset);
            val = decodeFromURL(val, charset);
            List<String> parameterValues = (List<String>) (params.containsKey(key) ? params.get(key)
                    : new ArrayList<String>());

            parameterValues.add(val);
            params.put(key, parameterValues);
        }
        return params;
    }

    public static String getQueryStringFromUrl(String url) {
        int i = url.indexOf('?');
        if (i < 0) {
            return "";
        }
        return url.substring(i + 1);
    }

    public static String removeQueryStringFromUrl(String url) {
        int i = url.indexOf('?');
        if (i < 0) {
            return url;
        }
        return url.substring(0, i);
    }

    public static String removeEndingSlashes(String url) {
        if (url == null || url.isEmpty()) {
            return "";
        }
        StringBuilder s = new StringBuilder(url);
        while (s.charAt(s.length() - 1) == '/') {
            s.deleteCharAt(s.length() - 1);
        }
        return s.toString();
    }

    public static String buildSubUrl(String url, String... subpath) {
        String query = URLUtils.getQueryStringFromUrl(url);
        url = URLUtils.removeQueryStringFromUrl(url);
        url = URLUtils.removeEndingSlashes(url);

        StringBuilder path = new StringBuilder(url);
        for (String part : subpath) {
            if (part != null && !part.isEmpty()) {
                if (!part.startsWith("/")) {
                    path.append("/");
                }
                path.append(part);
            }
        }
        url = path.toString();
        url = URLUtils.appendQueryToUrl(url, query);
        return url;
    }

    public static URL removeDefaultPort(final URL url) {
        if (url.getPort() != -1 && url.getDefaultPort() == url.getPort()) {
            try {
                return new URL(url.getProtocol(), url.getHost(), url.getFile());
            } catch (MalformedURLException e) {
                throw new IllegalStateException("Cannot remove default port part from URL <"
                        + url + ">", e);
            }
        }
        return url;
    }

    public static String removeDefaultPort(final String inurl) {
        try {
            return removeDefaultPort(new URL(inurl)).toExternalForm();
        } catch (MalformedURLException e) {
            throw new IllegalArgumentException("Illegal URL <" + inurl + ">", e);
        }
    }

}
