/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 24.03.2008
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.xml;

import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.io.StringWriter;
import java.io.Writer;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.n52.security.common.util.AbstractObjectPool;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.traversal.NodeFilter;

/**
 * Utility calls responsible for serializing xml documents to output streams.
 * <p/>
 * Usage:
 * <p/>
 * Use DOMSerializer.createNew() to create a new instance of a DOMSerializier and then use one of the serialize methods.
 * <p/>
 * A DOMSerializer shall be never hold as member variable, please create them only local.
 *
 * @author Marko Reiprecht Created on 24.03.2008
 * @version $Revision: $
 */
public abstract class DOMSerializer {

    private static SimpleCopyTransformerPool copyTransformerPool = new SimpleCopyTransformerPool();

    public static DOMSerializer createNew() {
        return createNew(null);
    }

    public static DOMSerializer createNew(final DOMSerializerOptions options) {
        return new PoolingDOMSerializer(options == null ? DOMSerializerOptions.getDefaultOptions() : options,
                copyTransformerPool);
    }

    /**
     * Creates a new copy Transformer.
     *
     * @param options
     *            the options.
     * @return Transformer.
     * @throws DOMSerializerException
     *             if the copy transformer can't be created.
     */
    protected static Transformer createTransformer(final DOMSerializerOptions options) throws DOMSerializerException {
        try {
            TransformerFactory fac = TransformerFactory.newInstance();
            return applyTransformerOptions(fac.newTransformer(), options);
        } catch (TransformerConfigurationException e) {
            throw new DOMSerializerException(e);
        }
    }

    /**
     * Method applys the DOMSericalizerOptions to the copy tansformer.
     *
     * @param transformer
     *            the transformer.
     * @param options
     *            the options.
     * @return Transformer.
     */
    private static Transformer applyTransformerOptions(final Transformer transformer, final DOMSerializerOptions options) {
        transformer.setOutputProperty(OutputKeys.METHOD, "xml");
        transformer.setOutputProperty(OutputKeys.ENCODING, options.getEncoding());
        transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, options.isOmitXMLDeclaration() ? "yes" : "no");
        transformer.setOutputProperty(OutputKeys.INDENT, options.isIndent() ? "yes" : "no");
        transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount",
                String.valueOf(options.getIndentAmount()));
        return transformer;
    }

    /**
     * Removes the indent text nodes from the dom tree.
     * <p/>
     * The semantic of this "cleanup" method is to remove any text nodes, which has element silblings. So if the xml
     * schemata contains nodes, which support "mixed-content" this method removes the mixed-content text nodes. Please
     * use this method only in Situations where you know that the algorithm works.
     *
     * @param document
     *            the document.
     * @return Document.
     */
    public static Document removeIndentTextNodes(final Document document) {
        new DocumentTraverser().traverseDepthFirst(document.getDocumentElement(), new DocumentVisitorAdapter() {
            public void visit(final Node node) {
                check(node.getPreviousSibling(), node);
                check(node.getNextSibling(), node);
            }

            private void check(final Node silb, final Node node) {
                if (silb != null && Node.ELEMENT_NODE == silb.getNodeType()) {
                    node.getParentNode().removeChild(node);
                }
            }

            public short acceptNode(final Node n) {
                switch (n.getNodeType()) {
                case Node.TEXT_NODE:
                    return NodeFilter.FILTER_ACCEPT;
                default:
                    return NodeFilter.FILTER_SKIP;
                }
            }
        });
        return document;
    }

    /**
     * Serializes the xml document to the writer.
     *
     * @throws DOMSerializerException
     *             if error occurs.
     */
    public void serialize(final Document document, final Writer out) throws DOMSerializerException {
        serialize((Node) document, out);
    }

    /**
     * Serializes the document to the output stream.
     *
     * @throws DOMSerializerException
     *             if error occurs.
     */
    public void serialize(final Document document, final OutputStream out) throws DOMSerializerException {
        serialize((Node) document, out);
    }

    /**
     * Converts the document to a byte[].
     *
     * @throws DOMSerializerException
     *             if error occurs.
     */
    public byte[] serializeToBytes(final Document document) throws DOMSerializerException {
        return serializeToBytes((Node) document);
    }

    /**
     * Converts the node to a byte[].
     *
     * @throws DOMSerializerException
     *             if error occurs.
     */
    public byte[] serializeToBytes(final Node node) throws DOMSerializerException {
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        serialize(node, bout);
        return bout.toByteArray();
    }

    /**
     * Serializes the node to the output stream.
     *
     * @throws DOMSerializerException
     *             if error occurs.
     */
    public abstract void serialize(Node node, OutputStream out) throws DOMSerializerException;

    /**
     * Converts the document to a String.
     *
     * @throws DOMSerializerException
     *             if error occurs.
     */
    public String serializeToString(final Document document) throws DOMSerializerException {
        return serializeToString((Node) document);
    }

    /**
     * Converts the node to a String.
     *
     * @throws DOMSerializerException
     *             if error occurs.
     */
    public String serializeToString(final Node node) throws DOMSerializerException {
        StringWriter writer = new StringWriter();
        serialize(node, writer);
        return writer.toString();
    }

    /**
     * Serializes the xml node to the writer.
     *
     * @throws DOMSerializerException
     *             if error occurs.
     */
    public abstract void serialize(Node node, Writer out) throws DOMSerializerException;

    // -------------------------- INNER CLASSES --------------------------

    /**
     * Uses a small transfomer pool to get a transformer for the xml document.
     *
     * @author Marko Reiprecht Created on 23.04.2008
     * @version $Revision: $
     */
    private static class PoolingDOMSerializer extends DOMSerializer {

        private DOMSerializerOptions m_options;
        private SimpleCopyTransformerPool m_pool;

        public PoolingDOMSerializer(final DOMSerializerOptions domSerializerOptions,
                final SimpleCopyTransformerPool pool) {
            m_options = domSerializerOptions;
            m_pool = pool;
        }

        public void serialize(final Node node, final Writer out) throws DOMSerializerException {
            Transformer transformer = m_pool.acquire(m_options);
            try {
                transformer.transform(new DOMSource(node), new StreamResult(out));
            } catch (TransformerException e) {
                throw new DOMSerializerException(e);
            } finally {
                m_pool.release(transformer, m_options);
            }
        }

        /**
         * @see DOMSerializer#serialize(Node, OutputStream)
         */
        public void serialize(final Node node, final OutputStream out) throws DOMSerializerException {
            Transformer transformer = m_pool.acquire(m_options);
            try {
                transformer.transform(new DOMSource(node), new StreamResult(out));
            } catch (TransformerException e) {
                throw new DOMSerializerException(e);
            } finally {
                m_pool.release(transformer, m_options);
            }
        }
    }

    private static class SimpleCopyTransformerPool extends AbstractObjectPool<Transformer, DOMSerializerOptions> {

        public SimpleCopyTransformerPool() {
            super(5);
        }

        @Override
        public Transformer create(DOMSerializerOptions key) {
            return DOMSerializer.createTransformer(key);
        }
    }
}
