/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 04.02.2009
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.xml;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;

import org.w3c.dom.Attr;
import org.w3c.dom.Comment;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 * Smart wrapper around the Document.create* methods, which is desinged to support the creation of
 * DOM structures within source code.
 *
 * @author Marko Reiprecht
 */
public class XMLBuilder {

    private Document m_sourceDocument;
    private SimpleNamespaceContext m_namespaces;

    private XMLBuilder(final Document source) {
        if (source == null) {
            throw new IllegalArgumentException("Parameter <source> must not null!");
        }
        m_sourceDocument = source;
        m_namespaces = new SimpleNamespaceContext();
    }

    public static XMLBuilder withNewDocument() {
        return new XMLBuilder(DOMParser.createEmptyDocument());
    }

    public static XMLBuilder withDocument(final Document source) {
        return new XMLBuilder(source);
    }

    public Document getDocument() {
        return m_sourceDocument;
    }

    public SimpleNamespaceContext getNamespaces() {
        return m_namespaces;
    }

    public XMLBuilder appendToRoot(XMLElement element) {
        if (m_sourceDocument.getDocumentElement() != null) {
            m_sourceDocument.getDocumentElement().appendChild(element.toW3C());
        } else {
            m_sourceDocument.appendChild(element.toW3C());
        }
        return this;
    }

    private Node find(final String xpath) {
        Node node = XMLPathCtx.createNew().addNamespaces(m_namespaces).findIn(m_sourceDocument).node(xpath).get();
        if (node == null) {
            throw new IllegalArgumentException("XPath <" + xpath + "> must resolve to an node");
        }
        return (Element) node;
    }

    public XMLBuilder appendTo(String xpath, XMLElement element) {
        find(xpath).appendChild(element.toW3C());
        return this;
    }

    public XMLBuilder replaceNodeAt(String xpath, XMLElement element) {
        Node target = find(xpath);
        target.getParentNode().replaceChild(element.toW3C(), target);
        return this;
    }

    public XMLBuilder appendBefore(String xpath, XMLElement element) {
        Node target = find(xpath);
        target.getParentNode().insertBefore(element.toW3C(), target);
        return this;
    }

    public XMLBuilder appendAfter(String xpath, XMLElement element) {
        Node target = find(xpath);
        Node before = target.getNextSibling();
        if (before != null) {
            target.getParentNode().insertBefore(element.toW3C(), before);
        } else {
            target.getParentNode().appendChild(element.toW3C());
        }
        return this;
    }

    public XMLBuilder namespace(final String prefix, final String namespace) {
        m_namespaces.add(prefix, namespace);
        return this;
    }

    public XMLElement element(final String qname, Attr... attributes) {
        return element(PrefixedName.toQName(qname, m_namespaces), attributes);
    }

    public XMLElement element(QName qname, Attr... attributes) {
        Element elem = m_sourceDocument.createElementNS(qname.getNamespaceURI(), qname.getLocalPart());
        return new XMLElement(elem).attributes(attributes);
    }

    public Text text(String content) {
        if (content == null) {
            content = "";
        }
        return m_sourceDocument.createTextNode(content);
    }

    public Comment comment(String comment) {
        if (comment == null) {
            comment = "";
        }
        return m_sourceDocument.createComment(comment);
    }

    public Attr attribute(final String qname, final String value) {
        QName qn = PrefixedName.toQName(qname, m_namespaces);
        Attr attr;
        if (XMLConstants.NULL_NS_URI.equals(qn.getNamespaceURI())) {
            attr = m_sourceDocument.createAttribute(qname);
        } else {
            attr = m_sourceDocument.createAttributeNS(qn.getNamespaceURI(), qname);
        }
        attr.setValue(value);
        return attr;
    }

    public static class XMLElement {

        private Element m_elem;

        private XMLElement(Element elem) {
            m_elem = elem;
        }

        public XMLElement attribute(Attr attribute) {
            if (attribute == null) {
                return this;
            }
            if (XMLConstants.NULL_NS_URI.equals(attribute.getNamespaceURI())) {
                m_elem.setAttributeNode(attribute);
            } else {
                m_elem.setAttributeNodeNS(attribute);
            }
            return this;
        }

        public XMLElement attributes(Attr... attributes) {
            for (Attr attr : attributes) {
                attribute(attr);
            }
            return this;
        }

        public XMLElement childs(XMLElement... elems) {
            return children(elems);
        }

        public XMLElement childs(Element... elems) {
            return children(elems);
        }

        public XMLElement children(XMLElement... elems) {
            for (XMLElement elem : elems) {
                add(elem);
            }
            return this;
        }

        public XMLElement children(Element... elems) {
            for (Element elem : elems) {
                add(elem);
            }
            return this;
        }

        public XMLElement add(XMLElement elem) {
            if (elem == null) {
                return this;
            }
            return add(elem.toW3C());
        }

        public XMLElement add(Element elem) {
            if (elem == null) {
                return this;
            }
            if (m_elem.getOwnerDocument() != elem.getOwnerDocument()) {
                elem = (Element) m_elem.getOwnerDocument().importNode(elem, true);
            }
            m_elem.appendChild(elem);
            return this;
        }

        public XMLElement add(Text text) {
            if (text == null) {
                return this;
            }
            m_elem.appendChild(text);
            return this;
        }

        public XMLElement text(Text text) {
            return add(text);
        }

        public XMLElement add(Comment comment) {
            if (comment == null) {
                return this;
            }
            m_elem.appendChild(comment);
            return this;
        }

        public XMLElement comment(Comment comment) {
            return add(comment);
        }

        public Element toW3C() {
            return m_elem;
        }

    }

    private static class PrefixedName {

        private String prefix;
        private String localPart;

        public static PrefixedName parse(final String name) {
            String[] splitted = name.split(":");
            switch (splitted.length) {
            case 1:
                return new PrefixedName(XMLConstants.DEFAULT_NS_PREFIX, splitted[0]);
            case 2:
                return new PrefixedName(splitted[0], splitted[1]);
            default:
                throw new IllegalArgumentException("Parameter name must not have more than one colon");
            }
        }

        public static QName toQName(String name, SimpleNamespaceContext namespaces) {
            return parse(name).toQName(namespaces);
        }

        public QName toQName(SimpleNamespaceContext namespaces) {
            return new QName(namespaces.getNamespaceURI(getPrefix()), getLocalPart());
        }

        public PrefixedName(String prefix, String localPart) {
            this.prefix = prefix;
            this.localPart = localPart;
        }

        public String getPrefix() {
            return prefix;
        }

        public String getLocalPart() {
            return localPart;
        }
    }
}
